# Generates valid Said() strings based on sample user input strings
#
# Usage:     perl saidGen.exe
#            -  You will need a file called 'userInputs.txt' holding each user input string to be tested
#               example should be included with this script)
#            -  Also needed is vocab.000 file (vocab file of all words, exported from an existing game - the template
#               game should be fine)
#            -  Outputs said syntax to screen & to the output file saidStrings.txt 
#
# Examples:  perl saidGen.exe 
#
# To Do:     May be issues with 'special' character classes.
#
# Do whatever you want with this file, except remove my name.
# Jeremiah Nellis, nellisjp@hotmail.com
# Version 1.0 - Dec 26th 2010


  use strict;
  use warnings; 
  use Parse::Earley;
  use Data::Dumper;
  use Switch;

  sub loadVocab();
  sub loadSuffixes();
  sub parseString();
  sub generateSaid();
  sub by_pred_order;

  my $inputStr;
  my @inputTokens;
  my $parseFail;
  my $wordClassStr;
  my @vocab;
  my @suffixes;
  my $parser = new Parse::Earley;

  # This grammar does not match to tokens that exist in the vocab.900 - 
  # the Parse::Earley did not like tokens prefaced with numbers & following characters
  # however, anyone familiar with the vocab file should be able to recognize the 
  # translation of the tokens...
  my $grammar = q {

	startrule: C013
	F013: C013
	C013: B013
	    | B013 D013 O133
	    | B013 D013 A013
	    | B013 D013 O137
	    | B013 D013
	    | B013 D013 O133 D013
	    | B013 D013 D013
	    | B013 D013 E013
	    | B013 E013
	    | D013
	    | D013 E013
	    | B013 D013 O133 C013
	    | D013 C013
	D013: O131 O139
	    | O131 A013 O139
	    | A013 O139
	    | A013
	    | O139
	E013: O138 D013
	    | O136 D013
	    | O138 O136 D013
	    | O138 O136 O132
	    | B013
	    | B013 E013
	    | O138 D013 O133 D013
	O139: O130 O135 D013
	    | O130 O139
	    | O130
	B013: O136 B013
	    | O136 O130 B013
	    | O136
	    | O136 O133 F012
	    | O136 O133 O133 F012
	    | F012
	    | F012 O133
	    | F012 O133 O133
	    | F012 A013
	    | F012 O137
	    | O137 O137 B013
	    | O137 B013
	    | O137
	    | O137 O130 B013
	    | F012 O135 F012
	    | F012 F012 O135 F012
	A013: O132
	    | O132 A013
	O137: "400"  
	F012: "800"
	    | "200"
	O136: "200"
	O130: "100"
	O134: "080"
	O132: "040"
	O131: "020"
	O133: "010"
	O138: "008"
	O135: "004"

         };



  # Load up all the word definitions
  loadVocab();
  
  # Load the suffixes
  loadSuffixes();
  
  # Set the grammar rules to those specified in $grammar
  $parser->grammar($grammar);

  open (USERINPUT, "userInputs.txt") || die "Could not open userInputs.txt: $!\n";
  my $fileStr = <USERINPUT>;
  close USERINPUT; 
  
  open (OUTPUTFILE,">" . "saidStrings.txt") || die "Could not open saidStrings.txt: $!\n";

  
  while ($fileStr) {
      ($inputStr, $fileStr) = split(/\n/, $fileStr, 2); 
	  $inputStr =~ s/^\s+//;
	  if(length($inputStr) == 0) {
	     next;
	  }

	  print "Input string: " . $inputStr . "\n";
	  @inputTokens = split('\s+', $inputStr);

	  # Find the parts of speech for our input string
	  my $result = parseString();
	  
	  if($result eq "failure") {
	     print "\n";
	     next;
	  }

	  # Initialize the parser state
	  $parser->start('startrule');

	  while (1) {
		# Advance the parser state to the next token of $wordClassStr
		$parser->advance($wordClassStr);
		if ($parser->fails($wordClassStr, 'startrule')) {
		   print "Cannot parse input string with grammar\n";
		   $parseFail = 1;
		   last;
		}
		if ($parser->matches($wordClassStr, 'startrule')) {
		   $parseFail = 0;
		   last;
		}
	  }

	  if ($parseFail == 0) {
		# make said string
		generateSaid();
	  }
	  
	  print "\n";
  }

close OUTPUTFILE;
exit(0);
  
sub loadVocab() {

	my $vocabFile = "vocab.000";
	my $data;
	my $rest;
	my $index = 0;
	my $header;
	my $hex;
	my $hexVal;
	my $hexVal1;
	my $hexVal2;

	my $row;
	my $lastword;
	my $group1;
	my $group2;
	my $group3;
	my $word;
	my $wordClass1;
	my $wordClass2;
	my $wordClass3;


	# Read whole file in
	undef $/;
	open FILE, $vocabFile or die "Can't open $vocabFile - $!\n";
	# binmode FILE to supress conversion of line endings
	binmode FILE;
	$data = <FILE>;
	close FILE;
	# convert data to hex form
	$hex = unpack 'H*', $data;
	$rest = $hex;

	# Read the first 54 characters, skipping the header
	while ($index < 55) {
	   ($header, $rest) = split //, $rest, 2;
	   ($header, $rest) = split //, $rest, 2;
	   $index++;
	}

	while ($rest) {
	    ($hexVal1, $rest) = split //, $rest, 2;
		($hexVal2, $rest) = split //, $rest, 2;
		$hexVal = hex($hexVal1 . $hexVal2);

		    
		   # Any character above 128 indicates the end of a word, subtract 128 for the final letter
		   if ($hexVal - 128 > 0) {
			   $hexVal = chr($hexVal - int(128));
			   $word = $word . $hexVal;
			   
			   ($wordClass1, $rest) = split //, $rest, 2;
		       ($wordClass2, $rest) = split //, $rest, 2;
		       ($wordClass3, $rest) = split //, $rest, 2;
		   
		       ($group1, $rest) = split //, $rest, 2;
		       ($group2, $rest) = split //, $rest, 2;
		       ($group3, $rest) = split //, $rest, 2;

		       my %hashparts = (
				   word => $word,
				   wordClass => $wordClass1 . $wordClass2 . $wordClass3,
				   wordGroup => $group1 . $group2 . $group3
		       );
		       push @vocab, { %hashparts };


			   # Hold the last word so we can use the starting characters for the next word;
			   $lastword = $word;

			   # Read the next hex, convert to an integer and use that many characters from the start of the last word
			   (my $charCount1, $rest) = split //, $rest, 2;
			   unless (defined $rest) { last;}
			   (my $charCount2, $rest) = split //, $rest, 2;
			   	   
			   my $charCount = int($charCount1 . $charCount2);
			   #print "Copying " . $charCount . " characters from the last word\n";
			   if ($charCount > 0) {
			      $word = substr($lastword, 0, $charCount);
			   } else {
		      $word = undef;
		   } 		   
		  
			   $row++;
			   
	       } else {      
		   $word = $word . chr($hexVal);
	       }
		   

	}
}
# end loadVocab()

sub loadSuffixes() {
	my $suffixFile = do {
	   my $fh;
	   local $/ = undef;
	   open $fh, "vocab.901" or die "Could not open vocab.901 file!\n";
	   binmode $fh;
	   <$fh>;
	};

	# read off the file header (first 2 characters)
	(my $junk, $suffixFile) = split /\x2A/, $suffixFile, 2;

	while ($suffixFile) {
	    (my $postSuffixGroup, $suffixFile) = split /\x2A/, $suffixFile, 2;
	    (my $postSuffixWord, my $postSuffixClass) = split /\x00/, $postSuffixGroup, 2;  

	    $postSuffixClass = unpack 'H*', $postSuffixClass;
	    $postSuffixClass = substr($postSuffixClass,1,4);
	 
	    (my $preSuffixGroup, $suffixFile) = split /\x2A/, $suffixFile, 2;
	    (my $preSuffixWord, my $preSuffixClass) = split /\x00/, $preSuffixGroup, 2;  

	    $preSuffixClass = unpack 'H*', $preSuffixClass;
	    $preSuffixClass = substr($preSuffixClass,1,4);

	    my %hashparts = (
		 preSuffixWord => $preSuffixWord,
		 preSuffixClass => $preSuffixClass,
		 postSuffixWord => $postSuffixWord,
		 postSuffixClass => $postSuffixClass
	    );
	    push @suffixes, { %hashparts };
	}
}

sub parseString() {
    use Switch;

    #my $inputStr = shift;
    my $i;
    my $j;
	#my @words = split(' ', $inputStr);
	my @seps;

	my @parseTokens;
	my @output;
	my $saidString;
        my $impVerbFound = 0;
        my $convertedWordClass;
        my $bin;

	#
	#  Find out what part of speech each word is (from the dumped vocab.000 file)
	#

	$wordClassStr = "";
	

	# iterate through the input string
	for($i=0; $i<scalar(@inputTokens); $i++) {

			# look for a match of the current word to a word in the vocab
			my $foundWord = 0;
			for ($j=0; $j<scalar(@vocab); $j++) {
	   
					if($inputTokens[$i] eq $vocab[$j]{'word'}) {
					   $foundWord = 1;

                         $bin = unpack('B12', pack('H*',$vocab[$j]{'wordClass'}));
                         $convertedWordClass = simplifyWordClass($bin, $impVerbFound);
					   
					   # only one imperative verb allowed for an input
					   if ($convertedWordClass eq "800") {
					      $impVerbFound = 1;
					   }

					   if($wordClassStr eq "") {
						  $wordClassStr =  $convertedWordClass;
					   } else {

						  $wordClassStr = $wordClassStr . " " . $convertedWordClass;
					   }
					   last;
					} 
		   }
		  
		   if($foundWord == 0) {
  
				# take the inputted word, remove the suffix 
				# then check to see if the root word is in the vocab

				# iterate through the suffixes, looking for a match to our word
				my $k;
				for($k=0; $k<scalar(@suffixes); $k++) {
				  my $temp = $suffixes[$k]{'postSuffixWord'};
				  if($inputTokens[$i] =~ m/$temp$/) {
									 

					 # see if the root is in the vocab
					 my $m;
					 my $wordTest = substr($inputTokens[$i], 0, length($inputTokens[$i]) - length($temp)) 
																. $suffixes[$k]{'preSuffixWord'};
					 for($m=0; $m<scalar(@vocab); $m++)  {
						if($wordTest eq $vocab[$m]{'word'}) {
							$foundWord = 1; 
							print "Suffix substitution: '" . $inputTokens[$i] . "' is not in vocab. Replaced with '" 
								  . $vocab[$m]{'word'}. "'\n";

							# change the user input 
							$inputTokens[$i] = $vocab[$m]{'word'};

							$bin = unpack('B12', pack('H*',$suffixes[$k]{'postSuffixClass'}));
							$convertedWordClass = simplifyWordClass($bin, $impVerbFound);
						   
							# only one imperative verb allowed for an input
							if ($convertedWordClass eq "800") {
								  $impVerbFound = 1;
							}

							if($wordClassStr eq "") {
								 $wordClassStr =  $convertedWordClass;
							} else {
								 $wordClassStr = $wordClassStr . " " . $convertedWordClass;
							}
							last;
						}
					 }
	   
					 # if we found a suffix, skip looking throught the rest of them
					 if($foundWord == 1) {
						 last;
					 }

				  } 
				}			
		   
                # Root word (or suffixed) could not be found
		        if($foundWord == 0) {
		  	      print "The word '" . $inputTokens[$i] . "' is not in the vocabulary!\n";
			      return "failure";
                }
		  }
	}
    
	return "success";
}

sub generateSaid() {
   my @subject;
   my @suffix;
   my @predicate;
   my $i;
   my $subjectComplete = 0;
   my $switchSubjectandSuffix = 0;
   my $foundPrep = 0;

   #print "Generating said() string for " . $wordClassStr  . "\n";
   $wordClassStr =~ s/200/600/;   # Total hack - change ind. verb class # higher so a descending sort will work right
   #print $wordClassStr . "\n";

   my @wordClass = split('\s+', $wordClassStr);

   # Put all the words into 3 arrays
   for($i=0; $i<scalar(@wordClass); $i++) {
      if($wordClass[$i] eq "800" || $wordClass[$i] eq "600" || $wordClass[$i] eq "400" || $wordClass[$i] eq "010" || $wordClass[$i] eq "008") {   
        # imp. verb, ind. verb, adverb, prep
		
		 # if this is a preposition, we are moving into the suffix of a word
		 # for now, we will also treat the special class of 008 the same as a preposition
         if ($wordClass[$i] eq "010" || $wordClass[$i] eq "008") {
		    $foundPrep = 1;
            $subjectComplete = 1;
         }
		 
        push @predicate, {class => $wordClass[$i], inputIndex => $i};
      }
      if($wordClass[$i] eq "100" || $wordClass[$i] eq "020" || $wordClass[$i] eq "040") {   # noun, article, adjective
	     if($wordClass[$i] eq "020") {

		     if (@subject) {
               $subjectComplete = 1;
		       # if the subject is complete & we have run into an article,
		       # (but have not run into a preposition), we will set an indicator 
			   # that the suffix & subject need to be flip-flopped			   
			   if ($foundPrep == 0) {
			     $switchSubjectandSuffix = 1;
			   }
  
             }			 
   	    	 next;
		 }

		
         if($subjectComplete == 0) {
             push @subject, $i;
         } else {
             push @suffix, $i;
         }
      }
   }

   
   # Switch the subject & suffix, if necessary
   if($switchSubjectandSuffix == 1) {
     #print "-------------flipping subject & suffix!--------------\n";   
     my @tempArray = @subject;
     @subject = @suffix;
     @suffix = @tempArray;
   }
   


   #my @inputTokens = split('\s+', $inputStr);

   # sort predicate array descending by word class tuple
   @predicate = sort { $$b{'class'} cmp $$a{'class'} } @predicate;
   #print "@predicate\n";
   
   # move the last element from the subject & suffix array to the front
   my $noun = pop @subject;
   unshift @subject, $noun;

   if(@suffix) {
      my $noun = pop @suffix;
      unshift @suffix, $noun;
   }
   

   my $saidString;
   
   # generate the string
   $saidString = $inputTokens[$predicate[0]{'inputIndex'}];

   for($i=1; $i<scalar(@predicate); $i++) {
      $saidString = $saidString . "<" . $inputTokens[$predicate[$i]{'inputIndex'}];
   } 

   if(defined $subject[0]) {
      $saidString = $saidString . "/";
      $saidString = $saidString . $inputTokens[$subject[0]];
      for($i=1; $i<scalar(@subject); $i++) {
         $saidString = $saidString . "<" . $inputTokens[$subject[$i]];
      }
   }   

   if(@suffix) {
      $saidString = $saidString . "/";
      $saidString = $saidString . $inputTokens[$suffix[0]];
      for($i=1; $i<scalar(@suffix); $i++) {
         $saidString = $saidString . "<" . $inputTokens[$suffix[$i]];
      }
   }   

   my $outputString;
   $outputString = "(if(Said('";
   $outputString = $outputString . $saidString;
   $outputString = $outputString . "')) Print(\"";
   $outputString = $outputString . $saidString;
   $outputString = $outputString . '") )';   
   
   print "Word classes: " . $wordClassStr . "\n";
   print "Said string: " . $saidString;
   print OUTPUTFILE "// " . $inputStr . "\n";
   print OUTPUTFILE $outputString . "\n\n";
   print "\n";
   

}

sub simplifyWordClass($) {
   (my $input, my $impVerbFound) = @_;
   my $foundOne = 0;
   my $output = "";

   # if we already have an imperative word found, set the first bit to 0 
   # (basically, we will not take another imperative verb)
   if($impVerbFound == 1) {
      (my $token, $input) = split(//, $input, 2);
	  $output = $output . "0";
   }
   
   while($input) {
      (my $token, $input) = split(//, $input, 2);
	  
      if ($foundOne == 0) {
         $output = $output . $token; 
      } else {
         $output = $output . "0";
      }

      if ($token == 1) {
         $foundOne = 1;
      }

   }

   return unpack('H3', pack('B12', $output));
}

